@@ -24,12 +24,13 @@ char *argv0;
 
 #include "arg.h"
 
-#undef strlcpy
+#ifdef NEED_STRLCPY
 size_t strlcpy(char *, const char *, size_t);
+#endif /* NEED_STRLCPY */
 
 #define IRC_CHANNEL_MAX   200
-#define IRC_MSG_MAX       512 /* quaranteed to be <= than PIPE_BUF */
-#define PING_TIMEOUT      300
+#define IRC_MSG_MAX       512 /* guaranteed to be <= than PIPE_BUF */
+#define PING_TIMEOUT      600
 
 enum { TOK_NICKSRV = 0, TOK_USER, TOK_CMD, TOK_CHAN, TOK_ARG, TOK_TEXT, TOK_LAST };
 
@@ -53,8 +54,10 @@ static int       channel_open(Channel *);
 static void      channel_print(Channel *, const char *);
 static int       channel_reopen(Channel *);
 static void      channel_rm(Channel *);
+static void      cleanup(void);
 static void      create_dirtree(const char *);
 static void      create_filepath(char *, size_t, const char *, const char *, const char *);
+static void      die(const char *, ...);
 static void      ewritestr(int, const char *);
 static void      handle_channels_input(int, Channel *);
 static void      handle_server_output(int);
@@ -82,13 +85,24 @@ static char     _nick[32];         /* nickname at startup */
 static char     ircpath[PATH_MAX]; /* irc dir (-i) */
 static char     msg[IRC_MSG_MAX];  /* message buf used for communication */
 
+static void
+die(const char *fmt, ...)
+{
+	va_list ap;
+
+	va_start(ap, fmt);
+	vfprintf(stderr, fmt, ap);
+	va_end(ap);
+
+	cleanup();
+	exit(1);
+}
+
 static void
 usage(void)
 {
-	fprintf(stderr, "usage: %s <-s host> [-i <irc dir>] [-p <port>] "
-	        "[-u <sockname>] [-n <nick>] [-k <password>] "
-	        "[-f <fullname>]\n", argv0);
-	exit(1);
+	die("usage: %s -s host [-p port | -u sockname] [-i ircdir]\n"
+	    "	[-n nickname] [-f fullname] [-k env_pass]\n", argv0);
 }
 
 static void
@@ -101,12 +115,9 @@ ewritestr(int fd, const char *s)
 	for (off = 0; off < len; off += w) {
 		if ((w = write(fd, s + off, len - off)) == -1)
 			break;
-		off += w;
-	}
-	if (w == -1) {
-		fprintf(stderr, "%s: write: %s\n", argv0, strerror(errno));
-		exit(1);
 	}
+	if (w == -1)
+		die("%s: write: %s\n", argv0, strerror(errno));
 }
 
 /* creates directories bottom-up, if necessary */
@@ -139,9 +150,9 @@ static void
 channel_normalize_path(char *s)
 {
 	for (; *s; s++) {
-		if (isalpha(*s))
-			*s = tolower(*s);
-		else if (!isdigit(*s) && !strchr(".#&+!-", *s))
+		if (isalpha((unsigned char)*s))
+			*s = tolower((unsigned char)*s);
+		else if (!isdigit((unsigned char)*s) && !strchr(".#&+!-", *s))
 			*s = '_';
 	}
 }
@@ -162,6 +173,20 @@ channel_normalize_name(char *s)
 	*p = '\0';
 }
 
+static void
+cleanup(void)
+{
+	Channel *c, *tmp;
+
+	if (channelmaster)
+		channel_leave(channelmaster);
+
+	for (c = channels; c; c = tmp) {
+		tmp = c->next;
+		channel_leave(c);
+	}
+}
+
 static void
 create_filepath(char *filepath, size_t len, const char *path,
 	const char *channel, const char *suffix)
@@ -184,8 +209,7 @@ create_filepath(char *filepath, size_t len, const char *path,
 	return;
 
 error:
-	fprintf(stderr, "%s: path to irc directory too long\n", argv0);
-	exit(1);
+	die("%s: path to irc directory too long\n", argv0);
 }
 
 static int
@@ -229,11 +253,9 @@ channel_new(const char *name)
 	strlcpy(channelpath, name, sizeof(channelpath));
 	channel_normalize_path(channelpath);
 
-	if (!(c = calloc(1, sizeof(Channel)))) {
-		fprintf(stderr, "%s: calloc: %s\n", argv0, strerror(errno));
-		exit(1);
-	}
-	c->next = NULL;
+	if (!(c = calloc(1, sizeof(Channel))))
+		die("%s: calloc: %s\n", argv0, strerror(errno));
+
 	strlcpy(c->name, name, sizeof(c->name));
 	channel_normalize_name(c->name);
 
@@ -341,21 +363,17 @@ udsopen(const char *uds)
 	size_t len;
 	int fd;
 
-	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
-		fprintf(stderr, "%s: socket: %s\n", argv0, strerror(errno));
-		exit(1);
-	}
+	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
+		die("%s: socket: %s\n", argv0, strerror(errno));
 
 	sun.sun_family = AF_UNIX;
-	if (strlcpy(sun.sun_path, uds, sizeof(sun.sun_path)) >= sizeof(sun.sun_path)) {
-		fprintf(stderr, "%s: UNIX domain socket path truncation\n", argv0);
-		exit(1);
-	}
+	if (strlcpy(sun.sun_path, uds, sizeof(sun.sun_path)) >= sizeof(sun.sun_path))
+		die("%s: UNIX domain socket path truncation\n", argv0);
+
 	len = strlen(sun.sun_path) + 1 + sizeof(sun.sun_family);
-	if (connect(fd, (struct sockaddr *)&sun, len) == -1) {
-		fprintf(stderr, "%s: connect: %s\n", argv0, strerror(errno));
-		exit(1);
-	}
+	if (connect(fd, (struct sockaddr *)&sun, len) == -1)
+		die("%s: connect: %s\n", argv0, strerror(errno));
+
 	return fd;
 }
 
@@ -370,27 +388,23 @@ tcpopen(const char *host, const char *service)
 	hints.ai_flags = AI_NUMERICSERV; /* avoid name lookup for port */
 	hints.ai_socktype = SOCK_STREAM;
 
-	if ((e = getaddrinfo(host, service, &hints, &res))) {
-		fprintf(stderr, "%s: getaddrinfo: %s\n", argv0, gai_strerror(e));
-		exit(1);
-	}
+	if ((e = getaddrinfo(host, service, &hints, &res)))
+		die("%s: getaddrinfo: %s\n", argv0, gai_strerror(e));
 
 	for (rp = res; rp; rp = rp->ai_next) {
-		fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+		fd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
 		if (fd == -1)
 			continue;
-		if (connect(fd, res->ai_addr, res->ai_addrlen) == -1) {
+		if (connect(fd, rp->ai_addr, rp->ai_addrlen) == -1) {
 			close(fd);
 			fd = -1;
 			continue;
 		}
 		break; /* success */
 	}
-	if (fd == -1) {
-		fprintf(stderr, "%s: could not connect to %s:%s: %s\n",
+	if (fd == -1)
+		die("%s: could not connect to %s:%s: %s\n",
 			argv0, host, service, strerror(errno));
-		exit(1);
-	}
 
 	freeaddrinfo(res);
 	return fd;
@@ -459,15 +473,21 @@ proc_channels_input(int ircfd, Channel *c, char *buf)
 	char *p = NULL;
 	size_t buflen;
 
-	if (buf[0] != '/' && buf[0] != '\0') {
+	if (buf[0] == '\0')
+		return;
+	if (buf[0] != '/') {
 		proc_channels_privmsg(ircfd, c, buf);
 		return;
 	}
+
 	msg[0] = '\0';
+	if ((buflen = strlen(buf)) < 2)
+		return;
 	if (buf[2] == ' ' || buf[2] == '\0') {
-		buflen = strlen(buf);
 		switch (buf[1]) {
 		case 'j': /* join */
+			if (buflen < 3)
+				return;
 			if ((p = strchr(&buf[3], ' '))) /* password parameter */
 				*p = '\0';
 			if ((buf[3] == '#') || (buf[3] == '&') || (buf[3] == '+') ||
@@ -638,7 +658,7 @@ proc_server_cmd(int fd, char *buf)
 				argv[TOK_NICKSRV], argv[TOK_ARG],
 				argv[TOK_TEXT] ? argv[TOK_TEXT] : "");
 	} else if (!strcmp("NOTICE", argv[TOK_CMD])) {
-		snprintf(msg, sizeof(msg), "-!- \"%s\")",
+		snprintf(msg, sizeof(msg), "-!- \"%s\"",
 				argv[TOK_TEXT] ? argv[TOK_TEXT] : "");
 	} else if (!strcmp("PRIVMSG", argv[TOK_CMD])) {
 		snprintf(msg, sizeof(msg), "<%s> %s", argv[TOK_NICKSRV],
@@ -677,7 +697,14 @@ read_line(int fd, char *buf, size_t bufsiz)
 static void
 handle_channels_input(int ircfd, Channel *c)
 {
-	char buf[IRC_MSG_MAX];
+	/*
+	 * Do not allow to read this fully, since commands will be
+	 * prepended. It will result in too long lines sent to the
+	 * server.
+	 * TODO: Make this depend on the maximum metadata given by the
+	 * server at the beginning of the connection.
+	 */
+	char buf[IRC_MSG_MAX-64];
 
 	if (read_line(c->fdin, buf, sizeof(buf)) == -1) {
 		if (channel_reopen(c) == -1)
@@ -692,11 +719,9 @@ handle_server_output(int ircfd)
 {
 	char buf[IRC_MSG_MAX];
 
-	if (read_line(ircfd, buf, sizeof(buf)) == -1) {
-		fprintf(stderr, "%s: remote host closed connection: %s\n",
-		        argv0, strerror(errno));
-		exit(1);
-	}
+	if (read_line(ircfd, buf, sizeof(buf)) == -1)
+		die("%s: remote host closed connection: %s\n", argv0, strerror(errno));
+
 	fprintf(stdout, "%lu %s\n", (unsigned long)time(NULL), buf);
 	fflush(stdout);
 	proc_server_cmd(ircfd, buf);
@@ -745,11 +770,11 @@ run(int ircfd, const char *host)
 		if (r < 0) {
 			if (errno == EINTR)
 				continue;
-			fprintf(stderr, "%s: select: %s\n", argv0, strerror(errno));
-			exit(1);
+			die("%s: select: %s\n", argv0, strerror(errno));
 		} else if (r == 0) {
 			if (time(NULL) - last_response >= PING_TIMEOUT) {
 				channel_print(channelmaster, "-!- ii shutting down: ping timeout");
+				cleanup();
 				exit(2); /* status code 2 for timeout */
 			}
 			ewritestr(ircfd, ping_msg);
@@ -770,7 +795,6 @@ run(int ircfd, const char *host)
 int
 main(int argc, char *argv[])
 {
-	Channel *c, *tmp;
 	struct passwd *spw;
 	const char *key = NULL, *fullname = NULL, *host = "";
 	const char *uds = NULL, *service = "6667";
@@ -778,10 +802,9 @@ main(int argc, char *argv[])
 	int ircfd, r;
 
 	/* use nickname and home dir of user by default */
-	if (!(spw = getpwuid(getuid()))) {
-		fprintf(stderr, "%s: getpwuid: %s\n", argv0, strerror(errno));
-		exit(1);
-	}
+	if (!(spw = getpwuid(getuid())))
+		die("%s: getpwuid: %s\n", argv0, strerror(errno));
+
 	strlcpy(nick, spw->pw_name, sizeof(nick));
 	snprintf(prefix, sizeof(prefix), "%s/irc", spw->pw_dir);
 
@@ -810,7 +833,7 @@ main(int argc, char *argv[])
 	default:
 		usage();
 		break;
-	} ARGEND;
+	} ARGEND
 
 	if (!*host)
 		usage();
@@ -822,17 +845,13 @@ main(int argc, char *argv[])
 
 #ifdef __OpenBSD__
 	/* OpenBSD pledge(2) support */
-	if (pledge("stdio rpath wpath cpath dpath", NULL) == -1) {
-		fprintf(stderr, "%s: pledge: %s\n", argv0, strerror(errno));
-		exit(1);
-	}
+	if (pledge("stdio rpath wpath cpath dpath", NULL) == -1)
+		die("%s: pledge: %s\n", argv0, strerror(errno));
 #endif
 
 	r = snprintf(ircpath, sizeof(ircpath), "%s/%s", prefix, host);
-	if (r < 0 || (size_t)r >= sizeof(ircpath)) {
-		fprintf(stderr, "%s: path to irc directory too long\n", argv0);
-		exit(1);
-	}
+	if (r < 0 || (size_t)r >= sizeof(ircpath))
+		die("%s: path to irc directory too long\n", argv0);
 	create_dirtree(ircpath);
 
 	channelmaster = channel_add(""); /* master channel */
@@ -841,13 +860,7 @@ main(int argc, char *argv[])
 	loginuser(ircfd, host, fullname && *fullname ? fullname : nick);
 	setup();
 	run(ircfd, host);
-	if (channelmaster)
-		channel_leave(channelmaster);
-
-	for (c = channels; c; c = tmp) {
-		tmp = c->next;
-		channel_leave(c);
-	}
+	cleanup();
 
 	return 0;
 }
